using UnityEngine;
using System.Collections.Generic;

namespace FaceModel
{
    public class TongueControl : MVRScript
    {
        private List<JSONStorableFloat> Parameters;
        private JSONStorableFloat MouthOpen;
        private JSONStorableFloat TongueLength;
        private JSONStorableFloat TongueWidth;
        private JSONStorableFloat TongueCurl;
        private JSONStorableFloat TongueRoll;
        private JSONStorableFloat TongueSideSide;
        private JSONStorableFloat TongueTwist;

        private DAZMorph MouthOpenMorph;
        private DAZMorph TongueLengthMorph;
        private DAZMorph TongueInOutMorph;
        private DAZMorph TongueRaiseLowerMorph;
        private DAZMorph TongueCurlMorph;
        private DAZMorph TongueSideSideMorph;
        private DAZMorph TongueTwistMorph;
        private DAZMorph TongueRollMorph;
        private DAZMorph MouthNarrowLeftMorph;
        private DAZMorph MouthNarrowRightMorph;

        private Dictionary<string, JSONStorableFloat> DebugMorphsValue;
        private Dictionary<string, DAZMorph> DebugMorphs;
        private string[] DebugMorphsName = new string[]
        {
            // "Mouth Open Wide",
            // "Tongue Length",
            // "Tongue In-Out",
            // "Tongue Raise-Lower",
            // "Tongue Bend Tip",
            // "Tongue Curl",
            // "Tongue Side-Side",
            // "Tongue Twist",
            // "Tongue Roll 1",
            // "Mouth Narrow Left",
            // "Mouth Narrow Right",
        };

        public override void Init()
        {
            InitMorphs();
            InitParameters();
            InitUserInterface();
        }

        private void InitMorphs()
        {
            DAZCharacterSelector geometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
            GenerateDAZMorphsControlUI morphsControl = geometry?.morphsControlUI;
            if (morphsControl != null)
            {
                MouthOpenMorph = morphsControl.GetMorphByDisplayName("Mouth Open Wide");
                TongueLengthMorph = morphsControl.GetMorphByDisplayName("Tongue Length");
                TongueInOutMorph = morphsControl.GetMorphByDisplayName("Tongue In-Out");
                TongueRaiseLowerMorph = morphsControl.GetMorphByDisplayName("Tongue Raise-Lower");

                TongueCurlMorph = morphsControl.GetMorphByDisplayName("Tongue Curl");
                TongueSideSideMorph = morphsControl.GetMorphByDisplayName("Tongue Side-Side");
                TongueTwistMorph = morphsControl.GetMorphByDisplayName("Tongue Twist");
                TongueRollMorph = morphsControl.GetMorphByDisplayName("Tongue Roll 1");

                MouthNarrowLeftMorph = morphsControl.GetMorphByDisplayName("Mouth Narrow Left");
                MouthNarrowRightMorph = morphsControl.GetMorphByDisplayName("Mouth Narrow Right");


                DebugMorphsValue = new Dictionary<string, JSONStorableFloat>();
                DebugMorphs = new Dictionary<string, DAZMorph>();
                foreach (string morphName in DebugMorphsName)
                {
                    DAZMorph morph = morphsControl.GetMorphByDisplayName(morphName);
                    morph.ResetRange();
                    JSONStorableFloat morphValue = new JSONStorableFloat(morphName, morph.startValue, morph.min, morph.max, false);
                    morph.morphValue = morph.startValue;

                    morphValue.setCallbackFunction = (float val) => morph.morphValue = val;
                    CreateSlider(morphValue, true);

                    DebugMorphs.Add(morphName, morph);
                    DebugMorphsValue.Add(morphName, morphValue);
                }
            }
        }

        private void InitParameters()
        {
            Parameters = new List<JSONStorableFloat>();
            MouthOpen = RegisterParameter(new JSONStorableFloat("Mouth Open", 0, 0, 1f));
            TongueLength = RegisterParameter(new JSONStorableFloat("Tongue Length", 0, 0, 1f));
            TongueCurl = RegisterParameter(new JSONStorableFloat("Tongue Curl", 0, -1, 1));
            TongueSideSide = RegisterParameter(new JSONStorableFloat("Tongue Side-Side", 0, -1, 1));
            TongueTwist = RegisterParameter(new JSONStorableFloat("Tongue Twist", 0, -1, 1));
            TongueWidth = RegisterParameter(new JSONStorableFloat("Tongue Width", 0, -1, 1));
            TongueRoll = RegisterParameter(new JSONStorableFloat("Tongue Roll", .5f, 0, 1));
        }

        private void InitUserInterface()
        {
            CreateSlider(MouthOpen);
            CreateSlider(TongueLength);
            CreateSlider(TongueCurl);
            CreateSlider(TongueSideSide);
            CreateSlider(TongueTwist);
            CreateSlider(TongueWidth);
            CreateSlider(TongueRoll);

            UpdateMorphs();

            JSONStorableBool tongueCollision = containingAtom.GetStorableByID("TongueControl").GetBoolJSONParam("tongueCollision");
            JSONStorableBool tongueCollisionProxy = new JSONStorableBool("Tongue Collision", tongueCollision.val);
            tongueCollisionProxy.setCallbackFunction = (bool val) => tongueCollision.val = val;
            CreateToggle(tongueCollisionProxy, true);

            UIDynamicButton resetButton = CreateButton("Reset", true);
            resetButton.button.onClick.AddListener(ResetParameters);
        }

        private JSONStorableFloat RegisterParameter(JSONStorableFloat parameter)
        {
            RegisterFloat(parameter);
            Parameters.Add(parameter);
            parameter.setCallbackFunction += (float value) => UpdateMorphs();
            return parameter;
        }

        private void ResetParameters()
        {
            foreach (JSONStorableFloat parameter in Parameters)
            {
                parameter.valNoCallback = parameter.defaultVal;
            }
            UpdateMorphs();
        }

        private void UpdateMorphs()
        {
            float tongueTwist = TongueTwist.val * Mathf.Min(1, TongueLength.val * 4f);
            TongueTwistMorph.morphValue = tongueTwist;

            float minMouthOpen = Mathf.Min(.9f, TongueLength.val * (.9f / .3f));
            minMouthOpen += Mathf.Abs(tongueTwist) * .9f;
            float mouthOpen = Mathf.Max(minMouthOpen, MouthOpen.val * 1.5f);
            MouthOpenMorph.morphValue = mouthOpen;

            float tongueWidthLerp = Mathf.InverseLerp(0, .2f, TongueLength.val);
            float tongueWidth = tongueWidthLerp * TongueWidth.val;
            TongueLengthMorph.morphValue = (TongueLength.val) * .4f + tongueWidth * .09f;
            TongueInOutMorph.morphValue = tongueWidth + (1 - tongueWidthLerp);

            float tongueCurl = TongueCurl.val * Mathf.Min(1, TongueLength.val * TongueLength.val);
            TongueCurlMorph.morphValue = tongueCurl;

            TongueRollMorph.morphValue = TongueRoll.val;

            float tongueSideSide = TongueSideSide.val * Mathf.Min(1, TongueLength.val * 4f);
            TongueSideSideMorph.morphValue = tongueSideSide;
            MouthNarrowLeftMorph.morphValue = Mathf.Min(0, tongueSideSide);
            MouthNarrowRightMorph.morphValue = Mathf.Min(0, -tongueSideSide);

            float tongueRaise = minMouthOpen / .9f * .54f;
            tongueRaise += tongueCurl * .7f;
            tongueRaise += Mathf.Abs(tongueTwist) * -.1f;
            TongueRaiseLowerMorph.morphValue = tongueRaise;

            UpdateDebugMorphs();
        }

        private void UpdateDebugMorphs()
        {
            foreach (string morphName in DebugMorphsName)
            {
                DebugMorphsValue[morphName].valNoCallback = DebugMorphs[morphName].morphValue;
            }
        }

        public void OnDestroy()
        {
            ResetParameters();
        }
    }
}
